@@ -32,6 +32,7 @@ module Hledger.Read.Common (
   InputOpts (..),
   definputopts,
   rawOptsToInputOpts,
+  forecastPeriodFromRawOpts,
 
   -- * parsing utilities
   runTextParser,
@@ -48,6 +49,8 @@ module Hledger.Read.Common (
   journalCheckAccountsDeclared,
   journalCheckCommoditiesDeclared,
   journalCheckPayeesDeclared,
+  journalAddForecast,
+  journalAddAutoPostings,
   setYear,
   getYear,
   setDefaultCommodityAndStyle,
@@ -95,9 +98,10 @@ module Hledger.Read.Common (
   rawnumberp,
 
   -- ** comments
+  isLineCommentStart,
+  isSameLineCommentStart,
   multilinecommentp,
   emptyorcommentlinep,
-
   followingcommentp,
   transactioncommentp,
   postingcommentp,
@@ -106,8 +110,11 @@ module Hledger.Read.Common (
   bracketeddatetagsp,
 
   -- ** misc
-  singlespacedtextp,
-  singlespacedtextsatisfyingp,
+  noncommenttextp,
+  noncommenttext1p,
+  singlespacedtext1p,
+  singlespacednoncommenttext1p,
+  singlespacedtextsatisfying1p,
   singlespacep,
   skipNonNewlineSpaces,
   skipNonNewlineSpaces1,
@@ -123,13 +130,15 @@ import Prelude ()
 import "base-compat-batteries" Prelude.Compat hiding (fail, readFile)
 import Control.Applicative.Permutations (runPermutation, toPermutationWithDefault)
 import qualified "base-compat-batteries" Control.Monad.Fail.Compat as Fail (fail)
-import Control.Monad.Except (ExceptT(..), runExceptT, throwError)
+import Control.Monad.Except (ExceptT(..), liftEither, runExceptT, throwError)
 import Control.Monad.State.Strict hiding (fail)
 import Data.Bifunctor (bimap, second)
 import Data.Char (digitToInt, isDigit, isSpace)
 import Data.Decimal (DecimalRaw (Decimal), Decimal)
 import Data.Default (Default(..))
+import Data.Either (lefts, rights)
 import Data.Function ((&))
+import Data.Functor ((<&>))
 import Data.Functor.Identity (Identity)
 import "base-compat-batteries" Data.List.Compat
 import Data.List.NonEmpty (NonEmpty(..))
@@ -138,7 +147,7 @@ import qualified Data.Map as M
 import qualified Data.Semigroup as Sem
 import Data.Text (Text)
 import qualified Data.Text as T
-import Data.Time.Calendar (Day, fromGregorianValid, toGregorian)
+import Data.Time.Calendar (Day, addDays, fromGregorianValid, toGregorian)
 import Data.Time.LocalTime (LocalTime(..), TimeOfDay(..))
 import Data.Word (Word8)
 import System.Time (getClockTime)
@@ -150,8 +159,9 @@ import Text.Megaparsec.Custom
   finalErrorBundlePretty, parseErrorAt, parseErrorAtRegion)
 
 import Hledger.Data
+import Hledger.Query (Query(..), filterQuery, parseQueryTerm, queryEndDate, queryIsDate, simplifyQuery)
+import Hledger.Reports.ReportOptions (ReportOpts(..), queryFromFlags, rawOptsToReportOpts)
 import Hledger.Utils
-import Safe (headMay)
 import Text.Printf (printf)
 
 --- ** doctest setup
@@ -197,13 +207,13 @@ data InputOpts = InputOpts {
     ,mrules_file_       :: Maybe FilePath       -- ^ a conversion rules file to use (when reading CSV)
     ,aliases_           :: [String]             -- ^ account name aliases to apply
     ,anon_              :: Bool                 -- ^ do light anonymisation/obfuscation of the data
-    ,ignore_assertions_ :: Bool                 -- ^ don't check balance assertions
     ,new_               :: Bool                 -- ^ read only new transactions since this file was last read
     ,new_save_          :: Bool                 -- ^ save latest new transactions state for next time
     ,pivot_             :: String               -- ^ use the given field's value as the account name
+    ,forecast_          :: Maybe DateSpan       -- ^ span in which to generate forecast transactions
     ,auto_              :: Bool                 -- ^ generate automatic postings when journal is parsed
-    ,commoditystyles_   :: Maybe (M.Map CommoditySymbol AmountStyle) -- ^ optional commodity display styles affecting all files
-    ,strict_            :: Bool                 -- ^ do extra error checking (eg, all posted accounts are declared)
+    ,balancingopts_     :: BalancingOpts        -- ^ options for balancing transactions
+    ,strict_            :: Bool                 -- ^ do extra error checking (eg, all posted accounts are declared, no prices are inferred)
  } deriving (Show)
 
 instance Default InputOpts where def = definputopts
@@ -214,30 +224,58 @@ definputopts = InputOpts
     , mrules_file_       = Nothing
     , aliases_           = []
     , anon_              = False
-    , ignore_assertions_ = False
     , new_               = False
     , new_save_          = True
     , pivot_             = ""
+    , forecast_          = Nothing
     , auto_              = False
-    , commoditystyles_   = Nothing
+    , balancingopts_     = def
     , strict_            = False
     }
 
-rawOptsToInputOpts :: RawOpts -> InputOpts
-rawOptsToInputOpts rawopts = InputOpts{
-   -- files_             = listofstringopt "file" rawopts
-   mformat_           = Nothing
-  ,mrules_file_       = maybestringopt "rules-file" rawopts
-  ,aliases_           = listofstringopt "alias" rawopts
-  ,anon_              = boolopt "anon" rawopts
-  ,ignore_assertions_ = boolopt "ignore-assertions" rawopts
-  ,new_               = boolopt "new" rawopts
-  ,new_save_          = True
-  ,pivot_             = stringopt "pivot" rawopts
-  ,auto_              = boolopt "auto" rawopts
-  ,commoditystyles_   = Nothing
-  ,strict_            = boolopt "strict" rawopts
-  }
+-- | Parse an InputOpts from a RawOpts and the current date.
+-- This will fail with a usage error if the forecast period expression cannot be parsed.
+rawOptsToInputOpts :: RawOpts -> IO InputOpts
+rawOptsToInputOpts rawopts = do
+    d <- getCurrentDay
+
+    return InputOpts{
+       -- files_             = listofstringopt "file" rawopts
+       mformat_           = Nothing
+      ,mrules_file_       = maybestringopt "rules-file" rawopts
+      ,aliases_           = listofstringopt "alias" rawopts
+      ,anon_              = boolopt "anon" rawopts
+      ,new_               = boolopt "new" rawopts
+      ,new_save_          = True
+      ,pivot_             = stringopt "pivot" rawopts
+      ,forecast_          = forecastPeriodFromRawOpts d rawopts
+      ,auto_              = boolopt "auto" rawopts
+      ,balancingopts_     = def{ ignore_assertions_ = boolopt "ignore-assertions" rawopts
+                               , infer_prices_      = not noinferprice
+                               }
+      ,strict_            = boolopt "strict" rawopts
+      }
+  where noinferprice = boolopt "strict" rawopts || stringopt "args" rawopts == "balancednoautoconversion"
+
+-- | Get period expression from --forecast option.
+-- This will fail with a usage error if the forecast period expression cannot be parsed.
+forecastPeriodFromRawOpts :: Day -> RawOpts -> Maybe DateSpan
+forecastPeriodFromRawOpts d rawopts = case maybestringopt "forecast" rawopts of
+    Nothing -> Nothing
+    Just "" -> Just forecastspanDefault
+    Just str -> either (\e -> usageError $ "could not parse forecast period : "++customErrorBundlePretty e)
+                       (\(_,requestedspan) -> Just $ requestedspan `spanDefaultsFrom` forecastspanDefault) $
+                  parsePeriodExpr d $ stripquotes $ T.pack str
+  where
+    -- "They end on or before the specified report end date, or 180 days from today if unspecified."
+    mspecifiedend = dbg2 "specifieddates" $ queryEndDate False datequery
+    forecastendDefault = dbg2 "forecastendDefault" $ addDays 180 d
+    forecastspanDefault = DateSpan Nothing $ mspecifiedend <|> Just forecastendDefault
+    -- Do we really need to do all this work just to get the requested end date? This is duplicating
+    -- much of reportOptsToSpec.
+    ropts = rawOptsToReportOpts d rawopts
+    argsquery = lefts . rights . map (parseQueryTerm d) $ querystring_ ropts
+    datequery = simplifyQuery . filterQuery queryIsDate . And $ queryFromFlags ropts : argsquery
 
 --- ** parsing utilities
 
@@ -316,6 +354,8 @@ parseAndFinaliseJournal' parser iopts f txt = do
 --
 -- - save misc info and reverse transactions into their original parse order,
 --
+-- - add forecast transactions,
+--
 -- - evaluate balance assignments and balance each transaction,
 --
 -- - apply transaction modifiers (auto postings) if enabled,
@@ -325,52 +365,65 @@ parseAndFinaliseJournal' parser iopts f txt = do
 -- - infer transaction-implied market prices from transaction prices
 --
 journalFinalise :: InputOpts -> FilePath -> Text -> ParsedJournal -> ExceptT String IO Journal
-journalFinalise InputOpts{auto_,ignore_assertions_,commoditystyles_,strict_} f txt pj = do
-  t <- liftIO getClockTime
-  d <- liftIO getCurrentDay
-  let pj' =
-        pj{jglobalcommoditystyles=fromMaybe M.empty commoditystyles_}  -- save any global commodity styles
-        & journalAddFile (f, txt)  -- save the main file's info
-        & journalSetLastReadTime t -- save the last read time
-        & journalReverse -- convert all lists to the order they were parsed
+journalFinalise InputOpts{forecast_,auto_,balancingopts_,strict_} f txt pj = do
+    t <- liftIO getClockTime
+    d <- liftIO getCurrentDay
+    -- Infer and apply canonical styles for each commodity (or throw an error).
+    -- This affects transaction balancing/assertions/assignments, so needs to be done early.
+    liftEither $ checkAddAndBalance d <=< journalApplyCommodityStyles $
+        pj{jglobalcommoditystyles=fromMaybe mempty $ commodity_styles_ balancingopts_}  -- save any global commodity styles
+        & journalAddFile (f, txt)           -- save the main file's info
+        & journalSetLastReadTime t          -- save the last read time
+        & journalReverse                    -- convert all lists to the order they were parsed
+  where
+    checkAddAndBalance d j = do
+        when strict_ $ do
+          -- If in strict mode, check all postings are to declared accounts
+          journalCheckAccountsDeclared j
+          -- and using declared commodities
+          journalCheckCommoditiesDeclared j
 
-  -- If in strict mode, check all postings are to declared accounts
-  case if strict_ then journalCheckAccountsDeclared pj' else Right () of
-    Left e   -> throwError e
-    Right () ->
+        -- Add forecast transactions if enabled
+        journalAddForecast d forecast_ j
+        -- Add auto postings if enabled
+          & (if auto_ && not (null $ jtxnmodifiers j) then journalAddAutoPostings d balancingopts_ else pure)
+        -- Balance all transactions and maybe check balance assertions.
+          >>= journalBalanceTransactions balancingopts_
+        -- infer market prices from commodity-exchanging transactions
+          <&> journalInferMarketPricesFromTransactions
 
-      -- and using declared commodities
-      case if strict_ then journalCheckCommoditiesDeclared pj' else Right () of
-        Left e   -> throwError e
-        Right () ->
+journalAddAutoPostings :: Day -> BalancingOpts -> Journal -> Either String Journal
+journalAddAutoPostings d bopts =
+    -- Balance all transactions without checking balance assertions,
+    journalBalanceTransactions bopts{ignore_assertions_=True}
+    -- then add the auto postings
+    -- (Note adding auto postings after balancing means #893b fails;
+    -- adding them before balancing probably means #893a, #928, #938 fail.)
+    >=> journalModifyTransactions d
+    >=> journalApplyCommodityStyles
 
-          -- Infer and apply canonical styles for each commodity (or throw an error).
-          -- This affects transaction balancing/assertions/assignments, so needs to be done early.
-          case journalApplyCommodityStyles pj' of
-            Left e     -> throwError e
-            Right pj'' -> either throwError return $
-              pj''
-              & (if not auto_ || null (jtxnmodifiers pj'')
-                then
-                  -- Auto postings are not active.
-                  -- Balance all transactions and maybe check balance assertions.
-                  journalBalanceTransactions (not ignore_assertions_)
-                else \j -> do  -- Either monad
-                  -- Auto postings are active.
-                  -- Balance all transactions without checking balance assertions,
-                  j' <- journalBalanceTransactions False j
-                  -- then add the auto postings
-                  -- (Note adding auto postings after balancing means #893b fails;
-                  -- adding them before balancing probably means #893a, #928, #938 fail.)
-                  case journalModifyTransactions d j' of
-                    Left e -> throwError e
-                    Right j'' -> do
-                      -- then apply commodity styles once more, to style the auto posting amounts. (XXX inefficient ?)
-                      j''' <- journalApplyCommodityStyles j''
-                      -- then check balance assertions.
-                      journalBalanceTransactions (not ignore_assertions_) j'''
-                )
-            & fmap journalInferMarketPricesFromTransactions  -- infer market prices from commodity-exchanging transactions
+-- | Generate periodic transactions from all periodic transaction rules in the journal.
+-- These transactions are added to the in-memory Journal (but not the on-disk file).
+--
+-- The start & end date for generated periodic transactions are determined in
+-- a somewhat complicated way; see the hledger manual -> Periodic transactions.
+journalAddForecast :: Day -> Maybe DateSpan -> Journal -> Journal
+journalAddForecast _ Nothing              j = j
+journalAddForecast d (Just requestedspan) j = j{jtxns = jtxns j ++ forecasttxns}
+  where
+    forecasttxns =
+        map (txnTieKnot . transactionTransformPostings (postingApplyCommodityStyles $ journalCommodityStyles j))
+      . filter (spanContainsDate forecastspan . tdate)
+      . concatMap (`runPeriodicTransaction` forecastspan)
+      $ jperiodictxns j
+
+    -- "They can start no earlier than: the day following the latest normal transaction in the journal (or today if there are none)."
+    mjournalend   = dbg2 "journalEndDate" $ journalEndDate False j  -- ignore secondary dates
+    forecastbeginDefault = dbg2 "forecastbeginDefault" $ mjournalend <|> Just d
+
+    -- "They end on or before the specified report end date, or 180 days from today if unspecified."
+    forecastspan = dbg2 "forecastspan" $ dbg2 "forecastspan flag" requestedspan
+        `spanDefaultsFrom` DateSpan forecastbeginDefault (Just $ addDays 180 d)
 
 -- | Check that all the journal's transactions have payees declared with
 -- payee directives, returning an error message otherwise.
@@ -408,7 +461,7 @@ journalCheckAccountsDeclared j = sequence_ $ map checkacct $ journalPostings j
 -- | Check that all the commodities used in this journal's postings have been declared
 -- by commodity directives, returning an error message otherwise.
 journalCheckCommoditiesDeclared :: Journal -> Either String ()
-journalCheckCommoditiesDeclared j = 
+journalCheckCommoditiesDeclared j =
   sequence_ $ map checkcommodities $ journalPostings j
   where
     checkcommodities Posting{..} =
@@ -423,10 +476,8 @@ journalCheckCommoditiesDeclared j =
                           (linesPrepend "  " . (<>"\n") . textChomp $ showTransaction t)
       where
         mfirstundeclaredcomm =
-          headMay $ filter (not . (`elem` cs)) $ catMaybes $
-          (acommodity . baamount <$> pbalanceassertion) :
-          (map (Just . acommodity) . filter (/= missingamt) $ amounts pamount)
-        cs = journalCommoditiesDeclared j
+          find (`M.notMember` jcommodities j) . map acommodity $
+          (maybe id ((:) . baamount) pbalanceassertion) . filter (/= missingamt) $ amountsRaw pamount
 
 
 setYear :: Year -> JournalParser m ()
@@ -535,9 +586,11 @@ codep = option "" $ do
   char ')' <?> "closing bracket ')' for transaction code"
   pure code
 
+-- | Parse possibly empty text until a semicolon or newline.
+-- Whitespace is preserved (for now - perhaps helps preserve alignment 
+-- of same-line comments ?).
 descriptionp :: TextParser m Text
-descriptionp = takeWhileP Nothing (not . semicolonOrNewline)
-  where semicolonOrNewline c = c == ';' || c == '\n'
+descriptionp = noncommenttextp <?> "description"
 
 --- *** dates
 
@@ -697,19 +750,32 @@ modifiedaccountnamep = do
 -- It should have required parts to start with an alphanumeric;
 -- for now it remains as-is for backwards compatibility.
 accountnamep :: TextParser m AccountName
-accountnamep = singlespacedtextp
+accountnamep = singlespacedtext1p
 
+-- | Parse possibly empty text, including whitespace, 
+-- until a comment start (semicolon) or newline.
+noncommenttextp :: TextParser m T.Text
+noncommenttextp = takeWhileP Nothing (\c -> not $ isSameLineCommentStart c || isNewline c)
 
--- | Parse any text beginning with a non-whitespace character, until a
--- double space or the end of input.
--- TODO including characters which normally start a comment (;#) - exclude those ? 
-singlespacedtextp :: TextParser m T.Text
-singlespacedtextp = singlespacedtextsatisfyingp (const True)
+-- | Parse non-empty text, including whitespace, 
+-- until a comment start (semicolon) or newline.
+noncommenttext1p :: TextParser m T.Text
+noncommenttext1p = takeWhile1P Nothing (\c -> not $ isSameLineCommentStart c || isNewline c)
 
--- | Similar to 'singlespacedtextp', except that the text must only contain
--- characters satisfying the given predicate.
-singlespacedtextsatisfyingp :: (Char -> Bool) -> TextParser m T.Text
-singlespacedtextsatisfyingp pred = do
+-- | Parse non-empty, single-spaced text starting and ending with non-whitespace,
+-- until a double space or newline.
+singlespacedtext1p :: TextParser m T.Text
+singlespacedtext1p = singlespacedtextsatisfying1p (const True)
+
+-- | Parse non-empty, single-spaced text starting and ending with non-whitespace,
+-- until a comment start (semicolon), double space, or newline.
+singlespacednoncommenttext1p :: TextParser m T.Text
+singlespacednoncommenttext1p = singlespacedtextsatisfying1p (not . isSameLineCommentStart)
+
+-- | Parse non-empty, single-spaced text starting and ending with non-whitespace,
+-- where all characters satisfy the given predicate.
+singlespacedtextsatisfying1p :: (Char -> Bool) -> TextParser m T.Text
+singlespacedtextsatisfying1p pred = do
   firstPart <- partp
   otherParts <- many $ try $ singlespacep *> partp
   pure $! T.unwords $ firstPart : otherParts
@@ -729,7 +795,7 @@ spaceandamountormissingp :: JournalParser m MixedAmount
 spaceandamountormissingp =
   option missingmixedamt $ try $ do
     lift $ skipNonNewlineSpaces1
-    Mixed . (:[]) <$> amountp
+    mixedAmount <$> amountp
 
 -- | Parse a single-commodity amount, with optional symbol on the left
 -- or right, followed by, in any order: an optional transaction price,
@@ -855,7 +921,7 @@ amountp' s =
 
 -- | Parse a mixed amount from a string, or get an error.
 mamountp' :: String -> MixedAmount
-mamountp' = Mixed . (:[]) . amountp'
+mamountp' = mixedAmount . amountp'
 
 -- | Parse a minus or plus sign followed by zero or more spaces,
 -- or nothing, returning a function that negates or does nothing.
@@ -1182,13 +1248,27 @@ emptyorcommentlinep = do
   where
     skiplinecommentp :: TextParser m ()
     skiplinecommentp = do
-      satisfy $ \c -> c == ';' || c == '#' || c == '*'
-      void $ takeWhileP Nothing (\c -> c /= '\n')
+      satisfy isLineCommentStart
+      void $ takeWhileP Nothing (/= '\n')
       optional newline
       pure ()
 
 {-# INLINABLE emptyorcommentlinep #-}
 
+-- | Is this a character that, as the first non-whitespace on a line,
+-- starts a comment line ?
+isLineCommentStart :: Char -> Bool
+isLineCommentStart '#' = True
+isLineCommentStart '*' = True
+isLineCommentStart ';' = True
+isLineCommentStart _   = False
+
+-- | Is this a character that, appearing anywhere within a line,
+-- starts a comment ?
+isSameLineCommentStart :: Char -> Bool
+isSameLineCommentStart ';' = True
+isSameLineCommentStart _   = False
+
 -- A parser combinator for parsing (possibly multiline) comments
 -- following journal items.
 --
@@ -1560,7 +1640,7 @@ tests_Common = tests "Common" [
      assertParseError p "1.5555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555555" ""
 
   ,tests "spaceandamountormissingp" [
-     test "space and amount" $ assertParseEq spaceandamountormissingp " $47.18" (Mixed [usd 47.18])
+     test "space and amount" $ assertParseEq spaceandamountormissingp " $47.18" (mixedAmount $ usd 47.18)
     ,test "empty string" $ assertParseEq spaceandamountormissingp "" missingmixedamt
     -- ,test "just space" $ assertParseEq spaceandamountormissingp " " missingmixedamt  -- XXX should it ?
     -- ,test "just amount" $ assertParseError spaceandamountormissingp "$47.18" ""  -- succeeds, consuming nothing
