@@ -1,12 +1,7 @@
-{- |
--}
-
-{-# LANGUAGE CPP                 #-}
 {-# LANGUAGE NamedFieldPuns      #-}
 {-# LANGUAGE OverloadedStrings   #-}
 {-# LANGUAGE RecordWildCards     #-}
 {-# LANGUAGE ScopedTypeVariables #-}
-{-# LANGUAGE TupleSections       #-}
 
 module Hledger.Reports.BudgetReport (
   BudgetGoal,
@@ -20,36 +15,30 @@ module Hledger.Reports.BudgetReport (
   budgetReportAsText,
   budgetReportAsCsv,
   -- * Helpers
-  reportPeriodName,
   combineBudgetAndActual,
   -- * Tests
   tests_BudgetReport
 )
 where
 
+import Control.Applicative ((<|>))
 import Data.Decimal (roundTo)
 import Data.Default (def)
 import Data.HashMap.Strict (HashMap)
 import qualified Data.HashMap.Strict as HM
-import Data.List (nub, partition, transpose)
+import Data.List (find, partition, transpose)
 import Data.List.Extra (nubSort)
 import Data.Maybe (fromMaybe)
-#if !(MIN_VERSION_base(4,11,0))
-import Data.Monoid ((<>))
-#endif
-import Safe (headDef)
---import Data.List
---import Data.Maybe
-import qualified Data.Map as Map
 import Data.Map (Map)
+import qualified Data.Map as Map
+import qualified Data.Set as S
 import Data.Text (Text)
 import qualified Data.Text as T
 import qualified Data.Text.Lazy as TL
 import qualified Data.Text.Lazy.Builder as TB
 --import System.Console.CmdArgs.Explicit as C
 --import Lucid as L
-import Text.Tabular as T
-import Text.Tabular.AsciiWide as T
+import Text.Tabular.AsciiWide as Tab
 
 import Hledger.Data
 import Hledger.Utils
@@ -74,8 +63,8 @@ type BudgetDisplayCell = ((Text, Int), Maybe ((Text, Int), Maybe (Text, Int)))
 -- from all periodic transactions, calculate actual balance changes 
 -- from the regular transactions, and compare these to get a 'BudgetReport'.
 -- Unbudgeted accounts may be hidden or renamed (see journalWithBudgetAccountNames).
-budgetReport :: ReportSpec -> Bool -> DateSpan -> Journal -> BudgetReport
-budgetReport rspec assrt reportspan j = dbg4 "sortedbudgetreport" budgetreport
+budgetReport :: ReportSpec -> BalancingOpts -> DateSpan -> Journal -> BudgetReport
+budgetReport rspec bopts reportspan j = dbg4 "sortedbudgetreport" budgetreport
   where
     -- Budget report demands ALTree mode to ensure subaccounts and subaccount budgets are properly handled
     -- and that reports with and without --empty make sense when compared side by side
@@ -83,14 +72,14 @@ budgetReport rspec assrt reportspan j = dbg4 "sortedbudgetreport" budgetreport
     showunbudgeted = empty_ ropts
     budgetedaccts =
       dbg3 "budgetedacctsinperiod" $
-      nub $
-      concatMap expandAccountName $
+      S.fromList $
+      expandAccountNames $
       accountNamesFromPostings $
       concatMap tpostings $
       concatMap (`runPeriodicTransaction` reportspan) $
       jperiodictxns j
     actualj = journalWithBudgetAccountNames budgetedaccts showunbudgeted j
-    budgetj = journalAddBudgetGoalTransactions assrt ropts reportspan j
+    budgetj = journalAddBudgetGoalTransactions bopts ropts reportspan j
     actualreport@(PeriodicReport actualspans _ _) =
         dbg5 "actualreport" $ multiBalanceReport rspec{rsOpts=ropts{empty_=True}} actualj
     budgetgoalreport@(PeriodicReport _ budgetgoalitems budgetgoaltotals) =
@@ -108,9 +97,9 @@ budgetReport rspec assrt reportspan j = dbg4 "sortedbudgetreport" budgetreport
 -- Budget goal transactions are similar to forecast transactions except
 -- their purpose and effect is to define balance change goals, per account and period,
 -- for BudgetReport.
-journalAddBudgetGoalTransactions :: Bool -> ReportOpts -> DateSpan -> Journal -> Journal
-journalAddBudgetGoalTransactions assrt _ropts reportspan j =
-  either error' id $ journalBalanceTransactions assrt j{ jtxns = budgetts }  -- PARTIAL:
+journalAddBudgetGoalTransactions :: BalancingOpts -> ReportOpts -> DateSpan -> Journal -> Journal
+journalAddBudgetGoalTransactions bopts _ropts reportspan j =
+  either error' id $ journalBalanceTransactions bopts j{ jtxns = budgetts }  -- PARTIAL:
   where
     budgetspan = dbg3 "budget span" $ reportspan
     budgetts =
@@ -130,25 +119,20 @@ journalAddBudgetGoalTransactions assrt _ropts reportspan j =
 --    with a budget goal, so that only budgeted accounts are shown.
 --    This can be disabled by -E/--empty.
 --
-journalWithBudgetAccountNames :: [AccountName] -> Bool -> Journal -> Journal
-journalWithBudgetAccountNames budgetedaccts showunbudgeted j = 
-  dbg5With (("budget account names: "++).pshow.journalAccountNamesUsed) $ 
+journalWithBudgetAccountNames :: S.Set AccountName -> Bool -> Journal -> Journal
+journalWithBudgetAccountNames budgetedaccts showunbudgeted j =
+  dbg5With (("budget account names: "++).pshow.journalAccountNamesUsed) $
   j { jtxns = remapTxn <$> jtxns j }
   where
-    remapTxn = mapPostings (map remapPosting)
+    remapTxn = txnTieKnot . transactionTransformPostings remapPosting
+    remapPosting p = p { paccount = remapAccount $ paccount p, poriginal = poriginal p <|> Just p }
+    remapAccount a
+      | a `S.member` budgetedaccts = a
+      | Just p <- budgetedparent   = if showunbudgeted then a else p
+      | otherwise                  = if showunbudgeted then u <> acctsep <> a else u
       where
-        mapPostings f t = txnTieKnot $ t { tpostings = f $ tpostings t }
-        remapPosting p = p { paccount = remapAccount $ paccount p, poriginal = Just . fromMaybe p $ poriginal p }
-          where
-            remapAccount a
-              | hasbudget         = a
-              | hasbudgetedparent = if showunbudgeted then a else budgetedparent
-              | otherwise         = if showunbudgeted then u <> acctsep <> a else u
-              where
-                hasbudget = a `elem` budgetedaccts
-                hasbudgetedparent = not $ T.null budgetedparent
-                budgetedparent = headDef "" $ filter (`elem` budgetedaccts) $ parentAccountNames a
-                u = unbudgetedAccountName
+        budgetedparent = find (`S.member` budgetedaccts) $ parentAccountNames a
+        u = unbudgetedAccountName
 
 -- | Combine a per-account-and-subperiod report of budget goals, and one
 -- of actual change amounts, into a budget performance report.
@@ -206,7 +190,7 @@ combineBudgetAndActual ropts j
     sortedrows :: [BudgetReportRow] = sortRowsLike (mbrsorted unbudgetedrows ++ mbrsorted rows') rows
       where
         (unbudgetedrows, rows') = partition ((==unbudgetedAccountName) . prrFullName) rows
-        mbrsorted = map prrFullName . sortRows ropts j . map (fmap $ fromMaybe 0 . fst)
+        mbrsorted = map prrFullName . sortRows ropts j . map (fmap $ fromMaybe nullmixedamt . fst)
         rows = rows1 ++ rows2
 
     -- TODO: grand total & average shows 0% when there are no actual amounts, inconsistent with other cells
@@ -244,7 +228,7 @@ budgetReportAsText ropts@ReportOpts{..} budgetr = TB.toLazyText $
 
     displayCell (actual, budget) = (showamt actual', budgetAndPerc <$> budget)
       where
-        actual' = fromMaybe 0 actual
+        actual' = fromMaybe nullmixedamt actual
         budgetAndPerc b = (showamt b, showper <$> percentage actual' b)
         showamt = (\(WideBuilder b w) -> (TL.toStrict $ TB.toLazyText b, w)) . showMixedAmountB oneLine{displayColour=color_, displayMaxWidth=Just 32}
         showper p = let str = T.pack (show $ roundTo 0 p) in (str, T.length str)
@@ -280,15 +264,15 @@ budgetReportAsText ropts@ReportOpts{..} budgetr = TB.toLazyText $
     -- - the goal is zero
     percentage :: Change -> BudgetGoal -> Maybe Percentage
     percentage actual budget =
-      case (maybecost $ normaliseMixedAmount actual, maybecost $ normaliseMixedAmount budget) of
-        (Mixed [a], Mixed [b]) | (acommodity a == acommodity b || amountLooksZero a) && not (amountLooksZero b)
+      case (costedAmounts actual, costedAmounts budget) of
+        ([a], [b]) | (acommodity a == acommodity b || amountLooksZero a) && not (amountLooksZero b)
             -> Just $ 100 * aquantity a / aquantity b
         _   -> -- trace (pshow $ (maybecost actual, maybecost budget))  -- debug missing percentage
                Nothing
       where
-        maybecost = case cost_ of
-            Cost   -> mixedAmountCost
-            NoCost -> id
+        costedAmounts = case cost_ of
+            Cost   -> amounts . mixedAmountCost
+            NoCost -> amounts
 
     maybetranspose | transpose_ = \(Table rh ch vals) -> Table ch rh (transpose vals)
                    | otherwise  = id
@@ -300,8 +284,8 @@ budgetReportAsTable
   (PeriodicReport spans rows (PeriodicReportRow _ coltots grandtot grandavg)) =
     addtotalrow $
     Table
-      (T.Group NoLine $ map Header accts)
-      (T.Group NoLine $ map Header colheadings)
+      (Tab.Group NoLine $ map Header accts)
+      (Tab.Group NoLine $ map Header colheadings)
       (map rowvals rows)
   where
     colheadings = map (reportPeriodName balancetype_ spans) spans
@@ -324,27 +308,6 @@ budgetReportAsTable
                                ++ [grandavg | average_ ropts && not (null coltots)]
                     ))
 
--- | Make a name for the given period in a multiperiod report, given
--- the type of balance being reported and the full set of report
--- periods. This will be used as a column heading (or row heading, in
--- a register summary report). We try to pick a useful name as follows:
---
--- - ending-balance reports: the period's end date
---
--- - balance change reports where the periods are months and all in the same year:
---   the short month name in the current locale
---
--- - all other balance change reports: a description of the datespan,
---   abbreviated to compact form if possible (see showDateSpan).
---
-reportPeriodName :: BalanceType -> [DateSpan] -> DateSpan -> T.Text
-reportPeriodName balancetype spans =
-  case balancetype of
-    PeriodChange -> if multiyear then showDateSpan else showDateSpanMonthAbbrev
-      where
-        multiyear = (>1) $ length $ nubSort $ map spanStartYear spans
-    _ -> maybe "" (showDate . prevday) . spanEnd
-
 -- XXX generalise this with multiBalanceReportAsCsv ?
 -- | Render a budget report as CSV. Like multiBalanceReportAsCsv,
 -- but includes alternating actual and budget amount columns.
