@@ -28,9 +28,11 @@ module Hledger.Reports.ReportOptions (
   reportOptsToggleStatus,
   simplifyStatuses,
   whichDateFromOpts,
-  journalSelectingAmountFromOpts,
+  journalApplyValuationFromOpts,
+  journalApplyValuationFromOptsWith,
+  mixedAmountApplyValuationAfterSumFromOptsWith,
+  valuationAfterSum,
   intervalFromRawOpts,
-  forecastPeriodFromRawOpts,
   queryFromFlags,
   transactionDateFn,
   postingDateFn,
@@ -42,20 +44,19 @@ module Hledger.Reports.ReportOptions (
   reportPeriodOrJournalStart,
   reportPeriodLastDay,
   reportPeriodOrJournalLastDay,
+  reportPeriodName
 )
 where
 
 import Control.Applicative ((<|>))
+import Control.Monad ((<=<))
 import Data.List.Extra (nubSort)
-import Data.Maybe (fromMaybe, isJust, mapMaybe)
+import Data.Maybe (fromMaybe, mapMaybe)
 import qualified Data.Text as T
 import Data.Time.Calendar (Day, addDays)
 import Data.Default (Default(..))
 import Safe (headMay, lastDef, lastMay, maximumMay)
 
-import System.Console.ANSI (hSupportsANSIColor)
-import System.Environment (lookupEnv)
-import System.IO (stdout)
 import Text.Megaparsec.Custom
 
 import Hledger.Data
@@ -118,6 +119,7 @@ data ReportOpts = ReportOpts {
     ,drop_           :: Int
     ,row_total_      :: Bool
     ,no_total_       :: Bool
+    ,show_costs_     :: Bool  -- ^ Whether to show costs for reports which normally don't show them
     ,pretty_tables_  :: Bool
     ,sort_amount_    :: Bool
     ,percent_        :: Bool
@@ -135,7 +137,6 @@ data ReportOpts = ReportOpts {
       --   Influenced by the --color/colour flag (cf CliOptions),
       --   whether stdout is an interactive terminal, and the value of
       --   TERM and existence of NO_COLOR environment variables.
-    ,forecast_       :: Maybe DateSpan
     ,transpose_      :: Bool
  } deriving (Show)
 
@@ -165,33 +166,34 @@ defreportopts = ReportOpts
     , drop_            = 0
     , row_total_       = False
     , no_total_        = False
+    , show_costs_      = False
     , pretty_tables_   = False
     , sort_amount_     = False
     , percent_         = False
     , invert_          = False
     , normalbalance_   = Nothing
     , color_           = False
-    , forecast_        = Nothing
     , transpose_       = False
     }
 
-rawOptsToReportOpts :: RawOpts -> IO ReportOpts
-rawOptsToReportOpts rawopts = do
-    d <- getCurrentDay
-    no_color <- isJust <$> lookupEnv "NO_COLOR"
-    supports_color <- hSupportsANSIColor stdout
+-- | Generate a ReportOpts from raw command-line input, given a day.
+-- This will fail with a usage error if it is passed
+-- - an invalid --format argument,
+-- - an invalid --value argument,
+-- - if --valuechange is called with a valuation type other than -V/--value=end.
+rawOptsToReportOpts :: Day -> RawOpts -> ReportOpts
+rawOptsToReportOpts d rawopts =
 
-    let colorflag    = stringopt "color" rawopts
-        formatstring = T.pack <$> maybestringopt "format" rawopts
+    let formatstring = T.pack <$> maybestringopt "format" rawopts
         querystring  = map T.pack $ listofstringopt "args" rawopts  -- doesn't handle an arg like "" right
         (costing, valuation) = valuationTypeFromRawOpts rawopts
 
-    format <- case parseStringFormat <$> formatstring of
-        Nothing         -> return defaultBalanceLineFormat
-        Just (Right x)  -> return x
-        Just (Left err) -> fail $ "could not parse format option: " ++ err
+        format = case parseStringFormat <$> formatstring of
+            Nothing         -> defaultBalanceLineFormat
+            Just (Right x)  -> x
+            Just (Left err) -> usageError $ "could not parse format option: " ++ err
 
-    return defreportopts
+    in defreportopts
           {period_      = periodFromRawOpts d rawopts
           ,interval_    = intervalFromRawOpts rawopts
           ,statuses_    = statusesFromRawOpts rawopts
@@ -214,15 +216,12 @@ rawOptsToReportOpts rawopts = do
           ,drop_        = posintopt "drop" rawopts
           ,row_total_   = boolopt "row-total" rawopts
           ,no_total_    = boolopt "no-total" rawopts
+          ,show_costs_  = boolopt "show-costs" rawopts
           ,sort_amount_ = boolopt "sort-amount" rawopts
           ,percent_     = boolopt "percent" rawopts
           ,invert_      = boolopt "invert" rawopts
           ,pretty_tables_ = boolopt "pretty-tables" rawopts
-          ,color_       = and [not no_color
-                              ,not $ colorflag `elem` ["never","no"]
-                              ,colorflag `elem` ["always","yes"] || supports_color
-                              ]
-          ,forecast_    = forecastPeriodFromRawOpts d rawopts
+          ,color_       = useColorOnStdout -- a lower-level helper
           ,transpose_   = boolopt "transpose" rawopts
           }
 
@@ -276,7 +275,7 @@ updateReportSpecWith f rspec = reportOptsToSpec (rsToday rspec) . f $ rsOpts rsp
 rawOptsToReportSpec :: RawOpts -> IO ReportSpec
 rawOptsToReportSpec rawopts = do
     d <- getCurrentDay
-    ropts <- rawOptsToReportOpts rawopts
+    let ropts = rawOptsToReportOpts d rawopts
     either fail return $ reportOptsToSpec d ropts
 
 accountlistmodeopt :: RawOpts -> AccountListMode
@@ -389,17 +388,6 @@ intervalFromRawOpts = lastDef NoInterval . collectopts intervalfromrawopt
       | n == "yearly"    = Just $ Years 1
       | otherwise = Nothing
 
--- | get period expression from --forecast option
-forecastPeriodFromRawOpts :: Day -> RawOpts -> Maybe DateSpan
-forecastPeriodFromRawOpts d opts =
-  case maybestringopt "forecast" opts
-  of
-    Nothing -> Nothing
-    Just "" -> Just nulldatespan
-    Just str ->
-      either (\e -> usageError $ "could not parse forecast period : "++customErrorBundlePretty e) (Just . snd) $ 
-      parsePeriodExpr d $ stripquotes $ T.pack str
-
 -- | Extract the interval from the parsed -p/--period expression.
 -- Return Nothing if an interval is not explicitly defined.
 extractIntervalOrNothing :: (Interval, DateSpan) -> Maybe Interval
@@ -436,6 +424,9 @@ reportOptsToggleStatus s ropts@ReportOpts{statuses_=ss}
 -- specified by -B/--cost, -V, -X/--exchange, or --value flags. It is
 -- allowed to combine -B/--cost with any other valuation type. If
 -- there's more than one valuation type, the rightmost flag wins.
+-- This will fail with a usage error if an invalid argument is passed
+-- to --value, or if --valuechange is called with a valuation type
+-- other than -V/--value=end.
 valuationTypeFromRawOpts :: RawOpts -> (Costing, Maybe ValuationType)
 valuationTypeFromRawOpts rawopts = (costing, valuation)
   where
@@ -493,13 +484,61 @@ flat_ = not . tree_
 -- depthFromOpts :: ReportOpts -> Int
 -- depthFromOpts opts = min (fromMaybe 99999 $ depth_ opts) (queryDepth $ queryFromOpts nulldate opts)
 
--- | Convert this journal's postings' amounts to cost using their
--- transaction prices, if specified by options (-B/--cost).
--- Maybe soon superseded by newer valuation code.
-journalSelectingAmountFromOpts :: ReportOpts -> Journal -> Journal
-journalSelectingAmountFromOpts opts = case cost_ opts of
-    Cost   -> journalToCost
-    NoCost -> id
+-- | Convert this journal's postings' amounts to cost and/or to value, if specified
+-- by options (-B/--cost/-V/-X/--value etc.). Strip prices if not needed. This
+-- should be the main stop for performing costing and valuation. The exception is
+-- whenever you need to perform valuation _after_ summing up amounts, as in a
+-- historical balance report with --value=end. valuationAfterSum will check for this
+-- condition.
+journalApplyValuationFromOpts :: ReportSpec -> Journal -> Journal
+journalApplyValuationFromOpts rspec j =
+    journalApplyValuationFromOptsWith rspec j priceoracle
+  where priceoracle = journalPriceOracle (infer_value_ $ rsOpts rspec) j
+
+-- | Like journalApplyValuationFromOpts, but takes PriceOracle as an argument.
+journalApplyValuationFromOptsWith :: ReportSpec -> Journal -> PriceOracle -> Journal
+journalApplyValuationFromOptsWith rspec@ReportSpec{rsOpts=ropts} j priceoracle =
+    journalMapPostings valuation $ costing j
+  where
+    valuation p = maybe id (postingApplyValuation priceoracle styles (periodEnd p) (rsToday rspec)) (value_ ropts) p
+    costing = case cost_ ropts of
+        Cost   -> journalToCost
+        NoCost -> id
+
+    -- Find the end of the period containing this posting
+    periodEnd  = addDays (-1) . fromMaybe err . mPeriodEnd . postingDate
+    mPeriodEnd = spanEnd <=< latestSpanContaining (historical : spans)
+    historical = DateSpan Nothing $ spanStart =<< headMay spans
+    spans = splitSpan (interval_ ropts) $ reportSpanBothDates j rspec
+    styles = journalCommodityStyles j
+    err = error "journalApplyValuationFromOpts: expected all spans to have an end date"
+
+-- | Select the Account valuation functions required for performing valuation after summing
+-- amounts. Used in MultiBalanceReport to value historical and similar reports.
+mixedAmountApplyValuationAfterSumFromOptsWith :: ReportOpts -> Journal -> PriceOracle
+                                              -> (DateSpan -> MixedAmount -> MixedAmount)
+mixedAmountApplyValuationAfterSumFromOptsWith ropts j priceoracle =
+    case valuationAfterSum ropts of
+      Just mc -> \span -> valuation mc span . costing
+      Nothing -> const id
+  where
+    valuation mc span = mixedAmountValueAtDate priceoracle styles mc (maybe err (addDays (-1)) $ spanEnd span)
+      where err = error "mixedAmountApplyValuationAfterSumFromOptsWith: expected all spans to have an end date"
+    costing = case cost_ ropts of
+        Cost   -> styleMixedAmount styles . mixedAmountCost
+        NoCost -> id
+    styles = journalCommodityStyles j
+
+-- | If the ReportOpts specify that we are performing valuation after summing amounts,
+-- return Just the commodity symbol we're converting to, otherwise return Nothing.
+-- Used for example with historical reports with --value=end.
+valuationAfterSum :: ReportOpts -> Maybe (Maybe CommoditySymbol)
+valuationAfterSum ropts = case value_ ropts of
+    Just (AtEnd mc) | valueAfterSum -> Just mc
+    _                               -> Nothing
+  where valueAfterSum = reporttype_  ropts == ValueChangeReport
+                     || balancetype_ ropts /= PeriodChange
+
 
 -- | Convert report options to a query, ignoring any non-flag command line arguments.
 queryFromFlags :: ReportOpts -> Query
@@ -519,8 +558,6 @@ queryFromFlags ReportOpts{..} = simplifyQuery $ And flagsq
 -- options or queries, or otherwise the earliest and latest transaction or
 -- posting dates in the journal. If no dates are specified by options/queries
 -- and the journal is empty, returns the null date span.
--- The boolean argument flags whether primary and secondary dates are considered
--- equivalently.
 reportSpan :: Journal -> ReportSpec -> DateSpan
 reportSpan = reportSpanHelper False
 
@@ -591,6 +628,26 @@ reportPeriodOrJournalLastDay :: ReportSpec -> Journal -> Maybe Day
 reportPeriodOrJournalLastDay rspec j = reportPeriodLastDay rspec <|> journalOrPriceEnd
   where
     journalOrPriceEnd = case value_ $ rsOpts rspec of
-        Just (AtEnd _) -> max (journalEndDate False j) lastPriceDirective
-        _              -> journalEndDate False j
+        Just (AtEnd _) -> max (journalLastDay False j) lastPriceDirective
+        _              -> journalLastDay False j
     lastPriceDirective = fmap (addDays 1) . maximumMay . map pddate $ jpricedirectives j
+
+-- | Make a name for the given period in a multiperiod report, given
+-- the type of balance being reported and the full set of report
+-- periods. This will be used as a column heading (or row heading, in
+-- a register summary report). We try to pick a useful name as follows:
+--
+-- - ending-balance reports: the period's end date
+--
+-- - balance change reports where the periods are months and all in the same year:
+--   the short month name in the current locale
+--
+-- - all other balance change reports: a description of the datespan,
+--   abbreviated to compact form if possible (see showDateSpan).
+reportPeriodName :: BalanceType -> [DateSpan] -> DateSpan -> T.Text
+reportPeriodName balancetype spans =
+  case balancetype of
+    PeriodChange -> if multiyear then showDateSpan else showDateSpanMonthAbbrev
+      where
+        multiyear = (>1) $ length $ nubSort $ map spanStartYear spans
+    _ -> maybe "" (showDate . prevday) . spanEnd
